package com.vcarecity.cvs.service.impl;


import com.google.inject.Inject;
import com.google.inject.assistedinject.Assisted;
import com.vcarecity.cvs.core.AutoMapper;
import com.vcarecity.cvs.core.JavaMapColumn;
import com.vcarecity.cvs.core.ReflectionUtil;
import com.vcarecity.cvs.event.ExportEventListener;
import com.vcarecity.cvs.service.ResultHandlerService;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.csv.CSVFormat;
import org.apache.commons.csv.CSVPrinter;

import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.sql.Timestamp;
import java.time.*;
import java.time.format.DateTimeFormatter;
import java.util.List;
import java.util.stream.Collectors;

/**
 * @author Kerry on 19/11/21
 */

@Slf4j
public class CSVResultHandlerServiceImpl implements ResultHandlerService {

    private final CSVFormat csvFormat;

    private BufferedWriter bufferedWriter = null;
    private CSVPrinter writer = null;

    private final String filePattern;
    private int counter;
    private int fileNameIndex = 0;

    private File lastFile;

    private String table;
    private String[] lastHeader;

    /**
     * 分割文件行数
     */
    private static final int SPLIT_LINE_NUM = 200_0000;

    private final ExportEventListener eventListener;

    @Inject
    public CSVResultHandlerServiceImpl(ExportEventListener eventListener,
                                       @Assisted Object filePattern) {
        this.eventListener = eventListener;
        this.filePattern = (String) filePattern;
        this.csvFormat = CSVFormat.MYSQL;
        this.counter = 0;
    }

    private File getFile() {
        final int index = fileNameIndex;
        fileNameIndex++;

        lastFile = new File(filePattern + "-" + +index + ".csv");
        // call back
        eventListener.beforeExport(lastFile.getAbsolutePath());

        return lastFile;
    }


    private void checkFile() {
        synchronized (this) {
            if (counter % SPLIT_LINE_NUM == 0) {
                try {
                    if (lastFile != null) {
                        eventListener.exportSuccess(this.lastFile.getAbsolutePath(), table, lastHeader);
                    }
                    if (bufferedWriter != null) {
                        bufferedWriter.close();
                    }
                    if (writer != null) {
                        writer.close();
                    }
                    bufferedWriter = Files.newBufferedWriter(getFile().toPath());
                    writer = new CSVPrinter(bufferedWriter, csvFormat);
                } catch (IOException e) {
                    e.printStackTrace();
                    eventListener.exceptionCatch(e);
                }
            }
        }
    }

    @Override
    public <T> void resultHandler(String table, String[] header, List<T> dataList) throws Exception {
        this.table = table;
        this.lastHeader = header;
        checkFile();

        logger.debug("start write to csv file size = {}", dataList.size());

        long startTime = System.currentTimeMillis();

        Class<?> cls = Object.class;

        for (T t : dataList) {
            cls = t.getClass();
            Object[] row;
            if (t.getClass().isArray()) {
                row = (Object[]) t;
            } else if (List.class.isAssignableFrom(t.getClass())) {
                // noinspection rawtypes
                row = ((List) t).toArray();
            } else {
                // auto mapper
                //noinspection unchecked
                row = resultMapper(t, (Class<T>) t.getClass());
            }
            Object[] handlerRow = toCsvContent(row);
            writer.printRecord(handlerRow);
        }
        bufferedWriter.flush();

        counter += dataList.size();

        if (cls != Object.class) {
            // 输出Header
            final List<JavaMapColumn> fieldColumnList = ReflectionUtil.getFieldColumnList(cls);
            final List<String> collect = fieldColumnList.stream().map(JavaMapColumn::getColumn).collect(Collectors.toList());
            final String join = String.join(",", collect.toArray(new String[]{}));
            logger.debug("write to csv file usage time {} /ms. header = ({})", (System.currentTimeMillis() - startTime), join);
        } else {
            logger.debug("write to csv file usage time {} /ms. ", (System.currentTimeMillis() - startTime));
        }
    }


    protected <T> Object[] resultMapper(T t, Class<T> cls) {
        return AutoMapper.autoMapper(t, cls);
    }

    public void close() throws IOException {
        eventListener.exportSuccess(this.lastFile.getAbsolutePath(), table, this.lastHeader);
        bufferedWriter.close();
        writer.close();
    }

    private static Object[] toCsvContent(Object[] objects) {
        Object[] contents = new Object[objects.length];

        // 主要是对 时间格式的处理

        for (int i = 0; i < objects.length; i++) {
            final Object object = objects[i];
            if (object == null) {
                contents[i] = null;
            } else if (object instanceof Timestamp) {
                contents[i] = formatTimestamp((Timestamp) object);
            } else if (object instanceof java.sql.Date) {
                contents[i] = formatDate((java.sql.Date) object);
            } else if (object instanceof java.util.Date) {
                contents[i] = formatDate((java.util.Date) object);
            } else if (object instanceof LocalDateTime) {
                contents[i] = formatLocalDateTime((LocalDateTime) object);
            } else if (object instanceof LocalDate) {
                contents[i] = formatLocalDate((LocalDate) object);
            } else if (object instanceof LocalTime) {
                contents[i] = formatLocalTime((LocalTime) object);
            } else {
                contents[i] = object;
            }
        }
        return contents;
    }

    public static final DateTimeFormatter DATE_FORMATTER = DateTimeFormatter.ofPattern("yyyy/MM/dd");

    private static String formatLocalDate(LocalDate localDate) {
        return localDate.format(DATE_FORMATTER);
    }


    private static String formatDate(java.sql.Date date) {
        final LocalDate localDate = date.toLocalDate();
        return localDate.format(DATE_FORMATTER);
    }

    private static String formatDate(java.util.Date date) {
        final Instant instant = date.toInstant();
        final LocalDateTime of = LocalDateTime.ofInstant(instant, ZoneId.systemDefault());
        return of.format(DATE_TIME_FORMATTER);
    }

    public static final DateTimeFormatter DATE_TIME_FORMATTER = DateTimeFormatter.ofPattern("yyyy/MM/dd HH:mm:ss");

    private static String formatTimestamp(Timestamp timestamp) {
        final LocalDateTime localDateTime = timestamp.toLocalDateTime();
        return localDateTime.format(DATE_TIME_FORMATTER);
    }

    private static String formatLocalDateTime(LocalDateTime localDateTime) {
        return localDateTime.format(DATE_TIME_FORMATTER);
    }

    public static final DateTimeFormatter TIME_FORMATTER = DateTimeFormatter.ofPattern("HH:mm:ss");

    private static String formatLocalTime(LocalTime localTime) {
        return localTime.format(TIME_FORMATTER);
    }


}
