package com.vcarecity.cvs.service.impl;

import com.google.inject.Inject;
import com.google.inject.assistedinject.Assisted;
import com.vcarecity.cvs.core.ColumnUpdateMapper;
import com.vcarecity.cvs.properties.AppProperties;
import com.vcarecity.cvs.service.ResultHandlerService;
import com.vcarecity.cvs.service.TimerService;
import com.vcarecity.cvs.util.CreateDatasource;
import com.zaxxer.hikari.HikariDataSource;
import lombok.extern.slf4j.Slf4j;

import java.sql.*;
import java.util.List;
import java.util.concurrent.atomic.AtomicLong;

/**
 * @author Kerry on 19/11/21
 */


@Slf4j
public class MySQLResultStatHandlerServiceImpl implements ResultHandlerService, TimerService {

    private final HikariDataSource dataSource;

    private AtomicLong writeTime = new AtomicLong(0L);

    @Inject
    public MySQLResultStatHandlerServiceImpl(AppProperties properties, @Assisted Object sql) {
        this.dataSource = CreateDatasource.createDataSource(properties.getToDb());
    }

    @Override
    public void truncateTable(String table) {
        try (final Connection connection = dataSource.getConnection();
             final PreparedStatement statement = connection.prepareStatement("truncate table " + table)) {
            statement.execute();
        } catch (SQLException e) {
            e.printStackTrace();
        }
    }

    @Override
    public <T> void resultHandler(String table, String[] header, List<T> dataList) throws Exception {

        if (dataList == null || dataList.isEmpty()) {
            return;
        }
        long startTime = System.currentTimeMillis();


        final String[] newHeaders = ColumnUpdateMapper.updateColumnName(table, header);
        final String column = String.join(",", newHeaders);

        StringBuilder sb = new StringBuilder();
        sb.append("INSERT INTO ").append(table).append("(").append(String.join(",", column)).append(") VALUES ");

        for (T t : dataList) {
            sb.append("(");
            if (t.getClass().isArray()) {
                Object[] rows = (Object[]) t;
                for (int i = 0; i < header.length; i++) {
                    mapperRow(rows[i], sb);
                    if (i != header.length - 1) {
                        sb.append(",");
                    }
                }
            } else if (List.class.isAssignableFrom(t.getClass())) {
                List list = (List) t;
                for (int i = 0; i < list.size(); i++) {
                    mapperRow(list.get(i), sb);
                    if (i != header.length - 1) {
                        sb.append(",");
                    }
                }

            } else {
                //TODO auto mapper
            }
            sb.append("),");
        }
        String execSql = sb.deleteCharAt(sb.length() - 1).toString();

        logger.info("{}", execSql);

        try {
            final Connection connection = dataSource.getConnection();
            final Statement statement = connection.createStatement();
            statement.execute(execSql);
            statement.close();
            connection.close();
        } catch (SQLException e) {
            logger.warn("error sql: {}", execSql);
            throw e;
        }

        final long curUseTime = System.currentTimeMillis() - startTime;

        writeTime.addAndGet(curUseTime);

        logger.info("WRITE SUCCESS to mysql usage time {} /ms, size = {}", curUseTime, dataList.size());
    }

    private void mapperRow(Object row, StringBuilder sb) {
        if (row instanceof String) {
            sb.append("'").append(row).append("'");
        } else if (row instanceof Date) {
            sb.append("'").append(row).append("'");
        } else if (row instanceof Timestamp) {
            sb.append("'").append(row).append("'");
        } else {
            sb.append(row);
        }
    }


    @Override
    public long getUsageTime() {
        return this.writeTime.longValue();
    }
}
